# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cloudpickle
import numpy as np

from ...context import RunningMode
from ...core import Object, OBJECT_TYPE, OBJECT_CHUNK_TYPE
from ...serialize import KeyField, Int32Field, DictField, AnyField, BoolField
from ...tiles import TilesError
from ...tensor.core import TensorOrder
from ...utils import check_chunks_unknown_shape, tokenize
from ..operands import LearnOperand, LearnOperandMixin, OutputType


class TreeObject(Object):
    def fetch(self, session=None, **kw):
        result = self._data.fetch(session=session, **kw)
        return cloudpickle.loads(result) \
            if isinstance(result, bytes) else result


class TreeBase(LearnOperand, LearnOperandMixin):
    _input = KeyField('input')
    _leaf_size = Int32Field('leaf_size')
    _metric = AnyField('metric')

    _metric_params = DictField('metric_params')

    def __init__(self, leaf_size=None, metric=None,
                 metric_params=None, output_types=None, **kw):
        super().__init__(_leaf_size=leaf_size, _metric=metric,
                         _metric_params=metric_params,
                         _output_types=output_types, **kw)
        if self.output_types is None:
            self.output_types = [OutputType.object]

    @property
    def input(self):
        return self._input

    @property
    def leaf_size(self):
        return self._leaf_size

    @property
    def metric(self):
        return self._metric

    @property
    def metric_params(self):
        return self._metric_params

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, a):
        return self.new_tileable([a])

    @classmethod
    def tile(cls, op):
        check_chunks_unknown_shape(op.inputs, TilesError)

        # ball tree and kd tree requires the full data,
        # thus rechunk input tensor into 1 chunk
        inp = op.input.rechunk({ax: s for ax, s in enumerate(op.input.shape)})
        inp = inp._inplace_tile()
        out = op.outputs[0]

        chunk_op = op.copy().reset_key()
        kw = out.params
        kw['index'] = inp.chunks[0].index
        chunk = chunk_op.new_chunk([inp.chunks[0]], kws=[kw])

        new_op = op.copy()
        tileable_kw = out.params
        tileable_kw['nsplits'] = ((1,),)
        tileable_kw['chunks'] = [chunk]
        return new_op.new_tileables(op.inputs, kws=[tileable_kw])

    @classmethod
    def execute(cls, ctx, op):
        if op.gpu:  # pragma: no cover
            raise NotImplementedError('Does not support tree-based '
                                      'nearest neighbors on GPU')

        a = ctx[op.input.key]
        tree = cls._tree_type(
            a, op.leaf_size, metric=op.metric,
            **(op.metric_params or dict()))
        if ctx.running_mode in [RunningMode.local_cluster, RunningMode.distributed]:
            # for local cluster and distributed, pickle always
            ctx[op.outputs[0].key] = cloudpickle.dumps(tree)
        else:
            # otherwise, to be clear for local, just put into storage directly
            ctx[op.outputs[0].key] = tree


def _on_serialize_tree(tree):
    return cloudpickle.dumps(tree) if not hasattr(tree, 'key') else tree


def _on_deserialize_tree(ser):
    return cloudpickle.loads(ser) if isinstance(ser, bytes) else ser


class TreeQueryBase(LearnOperand, LearnOperandMixin):
    _input = KeyField('input')
    _tree = AnyField('tree', on_serialize=_on_serialize_tree,
                     on_deserialize=_on_deserialize_tree)
    _n_neighbors = Int32Field('n_neighbors')
    _return_distance = BoolField('return_distance')

    def __init__(self, tree=None, n_neighbors=None, return_distance=None,
                 output_types=None, **kw):
        super().__init__(_tree=tree, _n_neighbors=n_neighbors,
                         _return_distance=return_distance,
                         _output_types=output_types, **kw)
        if self.output_types is None:
            self.output_types = [OutputType.tensor] * self.output_limit

    @property
    def input(self):
        return self._input

    @property
    def tree(self):
        return self._tree

    @property
    def n_neighbors(self):
        return self._n_neighbors

    @property
    def return_distance(self):
        return self._return_distance

    @property
    def output_limit(self):
        return 2 if self._return_distance else 1

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]
        if isinstance(self._tree, (OBJECT_TYPE, OBJECT_CHUNK_TYPE)):
            self._tree = self._inputs[1]

    def _update_key(self):
        values = []
        for value in self._values_:
            if isinstance(value, self._tree_type):
                values.append(cloudpickle.dumps(value))
            else:
                values.append(value)
        self._obj_set('_key', tokenize(type(self).__name__, *values))
        return self

    def __call__(self, x):
        kws = []
        if self._return_distance:
            kws.append({'shape': (x.shape[0], self._n_neighbors),
                        'dtype': np.dtype(np.float64),
                        'order': x.order,
                        'type': 'distance'})
        kws.append({
            'shape': (x.shape[0], self._n_neighbors),
            'dtype': np.dtype(np.int64),
            'order': TensorOrder.C_ORDER,
            'type': 'indices'
        })
        inputs = [x]
        if isinstance(self._tree, OBJECT_TYPE):
            inputs.append(self._tree)
        return self.new_tileables(inputs, kws=kws, output_limit=len(kws))

    @classmethod
    def tile(cls, op):
        inp = op.input

        if inp.chunk_shape[1] != 1:
            check_chunks_unknown_shape([inp], TilesError)
            inp = inp.rechunk({1: inp.shape[1]})._inplace_tile()

        tree_chunk = None
        if isinstance(op.tree, OBJECT_TYPE):
            tree_chunk = op.tree.chunks[0]
        out_chunks = [[] for _ in range(len(op.outputs))]
        for chunk in inp.chunks:
            chunk_op = op.copy().reset_key()
            if tree_chunk is not None:
                chunk_op._tree = tree_chunk
            chunk_kws = []
            if op.return_distance:
                chunk_kws.append({
                    'shape': (chunk.shape[0], op.n_neighbors),
                    'dtype': np.dtype(np.float64),
                    'order': chunk.order,
                    'index': chunk.index,
                    'type': 'distance'
                })
            chunk_kws.append({
                'shape': (chunk.shape[0], op.n_neighbors),
                'dtype': np.dtype(np.int64),
                'order': TensorOrder.C_ORDER,
                'index': chunk.index,
                'type': 'indices'
            })
            chunk_inputs = [chunk]
            if tree_chunk is not None:
                chunk_inputs.append(tree_chunk)
            chunks = chunk_op.new_chunks(chunk_inputs, kws=chunk_kws,
                                         output_limit=len(chunk_kws))
            for cs, c in zip(out_chunks, chunks):
                cs.append(c)

        kws = [o.params for o in op.outputs]
        nsplits = list(inp.nsplits)
        nsplits[1] = (op.n_neighbors,)
        if op.return_distance:
            kws[0]['chunks'] = out_chunks[0]
            kws[0]['nsplits'] = tuple(nsplits)
        kws[-1]['chunks'] = out_chunks[-1]
        kws[-1]['nsplits'] = tuple(nsplits)
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=kws, output_limit=len(kws))

    @classmethod
    def execute(cls, ctx, op):
        if op.gpu:  # pragma: no cover
            raise NotImplementedError('Does not support tree-based '
                                      'nearest neighbors on GPU')

        x = ctx[op.input.key]
        if len(op.inputs) == 2:
            tree = ctx[op.tree.key]
        else:
            tree = op.tree
        tree = cloudpickle.loads(tree) if isinstance(tree, bytes) else tree
        ret = tree.query(x, op.n_neighbors, op.return_distance)
        if op.return_distance:
            ctx[op.outputs[0].key] = ret[0]
            ctx[op.outputs[1].key] = ret[1]
        else:
            ctx[op.outputs[0].key] = ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.linalg import LinAlgError

from ...serialize import KeyField, BoolField
from ... import opcodes as OperandDef
from ...utils import check_chunks_unknown_shape
from ...tiles import TilesError
from ..operands import TensorHasInput, TensorOperand, TensorOperandMixin
from ..datasource import tensor as astensor
from ..core import TensorOrder
from ..array_utils import as_same_device, device


class TensorCholesky(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.CHOLESKY

    _input = KeyField('input')
    _lower = BoolField('lower')

    def __init__(self, lower=None, dtype=None, **kw):
        super().__init__(_lower=lower, _dtype=dtype, **kw)

    @property
    def lower(self):
        return self._lower

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, a):
        return self.new_tensor([a], a.shape, order=TensorOrder.C_ORDER)

    @classmethod
    def tile(cls, op):
        from ..datasource.zeros import TensorZeros
        from ..base import TensorTranspose
        from ..utils import reverse_order
        from .dot import TensorDot
        from .solve_triangular import TensorSolveTriangular

        tensor = op.outputs[0]
        in_tensor = op.input
        check_chunks_unknown_shape([in_tensor], TilesError)
        if in_tensor.nsplits[0] != in_tensor.nsplits[1]:
            # all chunks on diagonal should be square
            nsplits = in_tensor.nsplits[0]
            in_tensor = in_tensor.rechunk([nsplits, nsplits])._inplace_tile()

        lower_chunks, upper_chunks = {}, {}
        for i in range(in_tensor.chunk_shape[0]):
            for j in range(in_tensor.chunk_shape[1]):
                if i < j:
                    lower_chunk = TensorZeros(dtype=tensor.dtype).new_chunk(
                        None, shape=(in_tensor.nsplits[0][i], in_tensor.nsplits[1][j]),
                        index=(i, j), order=tensor.order)
                    upper_chunk = TensorZeros(dtype=tensor.dtype).new_chunk(
                        None, shape=(in_tensor.nsplits[1][j], in_tensor.nsplits[0][i]),
                        index=(j, i), order=tensor.order)
                    lower_chunks[lower_chunk.index] = lower_chunk
                    upper_chunks[upper_chunk.index] = upper_chunk
                elif i == j:
                    target = in_tensor.cix[i, j]
                    if i > 0:
                        prev_chunks = []
                        for p in range(i):
                            a, b = lower_chunks[i, p], upper_chunks[p, j]
                            prev_chunk = TensorDot(dtype=tensor.dtype).new_chunk(
                                [a, b], shape=(a.shape[0], b.shape[1]), order=tensor.order)
                            prev_chunks.append(prev_chunk)

                        cholesky_fuse_op = TensorCholeskyFuse()
                        lower_chunk = cholesky_fuse_op.new_chunk([target] + prev_chunks,
                                                                 shape=target.shape, index=(i, j),
                                                                 order=tensor.order)
                    else:
                        lower_chunk = TensorCholesky(lower=True, dtype=tensor.dtype).new_chunk(
                            [target], shape=target.shape, index=(i, j), order=tensor.order)

                    upper_chunk = TensorTranspose(dtype=lower_chunk.dtype).new_chunk(
                        [lower_chunk], shape=lower_chunk.shape[::-1],
                        index=lower_chunk.index[::-1], order=reverse_order(lower_chunk.order))
                    lower_chunks[lower_chunk.index] = lower_chunk
                    upper_chunks[upper_chunk.index] = upper_chunk
                else:
                    target = in_tensor.cix[j, i]
                    if j > 0:
                        prev_chunks = []
                        for p in range(j):
                            a, b = lower_chunks[j, p], upper_chunks[p, i]
                            prev_chunk = TensorDot(dtype=tensor.dtype).new_chunk(
                                [a, b], shape=(a.shape[0], b.shape[1]), order=tensor.order)
                            prev_chunks.append(prev_chunk)
                        cholesky_fuse_op = TensorCholeskyFuse(by_solve_triangular=True)
                        upper_chunk = cholesky_fuse_op.new_chunk([target] + [lower_chunks[j, j]] + prev_chunks,
                                                                 shape=target.shape, index=(j, i),
                                                                 order=tensor.order)
                    else:
                        upper_chunk = TensorSolveTriangular(lower=True, dtype=tensor.dtype).new_chunk(
                            [lower_chunks[j, j], target], shape=target.shape,
                            index=(j, i), order=tensor.order)
                    lower_chunk = TensorTranspose(dtype=upper_chunk.dtype).new_chunk(
                        [upper_chunk], shape=upper_chunk.shape[::-1],
                        index=upper_chunk.index[::-1], order=reverse_order(upper_chunk.order))
                    lower_chunks[lower_chunk.index] = lower_chunk
                    upper_chunks[upper_chunk.index] = upper_chunk

        new_op = op.copy()
        if op.lower:
            return new_op.new_tensors(op.inputs, tensor.shape, order=tensor.order,
                                      chunks=list(lower_chunks.values()), nsplits=in_tensor.nsplits)
        else:
            return new_op.new_tensors(op.inputs, tensor.shape, order=tensor.order,
                                      chunks=list(upper_chunks.values()), nsplits=in_tensor.nsplits)

    @classmethod
    def execute(cls, ctx, op):
        chunk = op.outputs[0]
        (a,), device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True)

        with device(device_id):
            if xp is np:
                try:
                    import scipy.linalg

                    ctx[chunk.key] = scipy.linalg.cholesky(a, lower=op.lower)
                    return
                except ImportError:  # pragma: no cover
                    pass

            r = xp.linalg.cholesky(a)
            if not chunk.op.lower:
                r = r.T.conj()

            ctx[chunk.key] = r


class TensorCholeskyFuse(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.CHOLESKY_FUSE

    _by_solve_triangular = BoolField('by_solve_triangular')

    def __init__(self, by_solve_triangular=None, **kw):
        super().__init__(_by_solve_triangular=by_solve_triangular, **kw)

    @property
    def by_solve_triangular(self):
        return self._by_solve_triangular

    @classmethod
    def _execute_by_cholesky(cls, inputs):
        import scipy.linalg

        target = inputs[0]
        return scipy.linalg.cholesky((target - sum(inputs[1:])), lower=True)

    @classmethod
    def _execute_by_solve_striangular(cls, inputs):
        import scipy.linalg

        target = inputs[0]
        lower = inputs[1]
        return scipy.linalg.solve_triangular(lower, (target - sum(inputs[2:])), lower=True)

    @classmethod
    def execute(cls, ctx, op):
        inputs = [ctx[c.key] for c in op.inputs]
        if op.by_solve_triangular:
            ret = cls._execute_by_solve_striangular(inputs)
        else:
            ret = cls._execute_by_cholesky(inputs)
        ctx[op.outputs[0].key] = ret


def cholesky(a, lower=False):
    """
    Cholesky decomposition.

    Return the Cholesky decomposition, `L * L.H`, of the square matrix `a`,
    where `L` is lower-triangular and .H is the conjugate transpose operator
    (which is the ordinary transpose if `a` is real-valued).  `a` must be
    Hermitian (symmetric if real-valued) and positive-definite.  Only `L` is
    actually returned.

    Parameters
    ----------
    a : (..., M, M) array_like
        Hermitian (symmetric if all elements are real), positive-definite
        input matrix.
    lower : bool
        Whether to compute the upper or lower triangular Cholesky
        factorization.  Default is upper-triangular.

    Returns
    -------
    L : (..., M, M) array_like
        Upper or lower-triangular Cholesky factor of `a`.

    Raises
    ------
    LinAlgError
       If the decomposition fails, for example, if `a` is not
       positive-definite.

    Notes
    -----

    Broadcasting rules apply, see the `mt.linalg` documentation for
    details.

    The Cholesky decomposition is often used as a fast way of solving

    .. math:: A \\mathbf{x} = \\mathbf{b}

    (when `A` is both Hermitian/symmetric and positive-definite).

    First, we solve for :math:`\\mathbf{y}` in

    .. math:: L \\mathbf{y} = \\mathbf{b},

    and then for :math:`\\mathbf{x}` in

    .. math:: L.H \\mathbf{x} = \\mathbf{y}.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> A = mt.array([[1,-2j],[2j,5]])
    >>> A.execute()
    array([[ 1.+0.j,  0.-2.j],
           [ 0.+2.j,  5.+0.j]])
    >>> L = mt.linalg.cholesky(A, lower=True)
    >>> L.execute()
    array([[ 1.+0.j,  0.+0.j],
           [ 0.+2.j,  1.+0.j]])
    >>> mt.dot(L, L.T.conj()).execute() # verify that L * L.H = A
    array([[ 1.+0.j,  0.-2.j],
           [ 0.+2.j,  5.+0.j]])
    >>> A = [[1,-2j],[2j,5]] # what happens if A is only array_like?
    >>> mt.linalg.cholesky(A, lower=True).execute()
    array([[ 1.+0.j,  0.+0.j],
           [ 0.+2.j,  1.+0.j]])

    """
    a = astensor(a)

    if a.ndim != 2:
        raise LinAlgError(f'{a.ndim}-dimensional array given. '
                          'Tensor must be two-dimensional')
    if a.shape[0] != a.shape[1]:
        raise LinAlgError('Input must be square')

    cho = np.linalg.cholesky(np.array([[1, 2], [2, 5]], dtype=a.dtype))

    op = TensorCholesky(lower=lower, dtype=cho.dtype)
    return op(a)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes
from ...core import ExecutableTuple
from ...serialize import KeyField, Int32Field
from ...utils import calc_nsplits
from ..core import TensorOrder
from ..datasource import tensor as astensor
from ..operands import TensorOperand, TensorOperandMixin


class TensorGetShape(TensorOperand, TensorOperandMixin):
    _op_type_ = opcodes.GET_SHAPE

    _a = KeyField('a')
    _ndim = Int32Field('ndim')

    def __init__(self, pure_depends=None, a=None, ndim=None, dtype=None, **kw):
        super().__init__(_dtype=dtype, _a=a, _ndim=ndim,
                         _pure_depends=pure_depends, **kw)

    @property
    def a(self):
        return self._a

    @property
    def ndim(self):
        return self._ndim

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self._a is not None:
            self._a = self._inputs[0]

    @property
    def output_limit(self):
        return self._ndim

    def __call__(self, a):
        if not np.isnan(a.size):
            return ExecutableTuple([astensor(s) for s in a.shape])

        self._a = a
        kws = []
        for i in range(self.output_limit):
            kws.append({
                'shape': (),
                'dtype': np.dtype(np.intc),
                'order': TensorOrder.C_ORDER,
                'i': i
            })
        return ExecutableTuple(self.new_tensors([a], kws=kws))

    @classmethod
    def tile(cls, op):
        a = op.a
        outs = op.outputs

        chunk_op = TensorGetShape(pure_depends=[True] * len(a.chunks),
                                  ndim=op.ndim)
        chunk_kws = []
        for out in outs:
            params = out.params
            params['index'] = ()
            chunk_kws.append(params)
        chunks = chunk_op.new_chunks(a.chunks, kws=chunk_kws)

        kws = []
        for c, out in zip(chunks, outs):
            params = out.params
            params['chunks'] = [c]
            params['nsplits'] = ()
            kws.append(params)
        new_op = op.copy()
        return new_op.new_tensors(op.inputs, kws=kws,
                                  output_limit=op.output_limit)

    @classmethod
    def execute(cls, ctx, op):
        chunk_idx_tochunk_shapes = \
            {c.index: cm.chunk_shape for c, cm
             in zip(op.inputs, ctx.get_chunk_metas([c.key for c in op.inputs]))}
        nsplits = calc_nsplits(chunk_idx_tochunk_shapes)
        shape = tuple(sum(ns) for ns in nsplits)
        for o, s in zip(op.outputs, shape):
            ctx[o.key] = s


def shape(a):
    """
    Return the shape of a tensor.

    Parameters
    ----------
    a : array_like
        Input tensor.

    Returns
    -------
    shape : ExecutableTuple of tensors
        The elements of the shape tuple give the lengths of the
        corresponding array dimensions.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.shape(mt.eye(3)).execute()
    (3, 3)
    >>> mt.shape([[1, 2]]).execute()
    (1, 2)
    >>> mt.shape([0]).execute()
    (1,)
    >>> mt.shape(0).execute()
    ()

    >>> a = mt.array([(1, 2), (3, 4)], dtype=[('x', 'i4'), ('y', 'i4')])
    >>> mt.shape(a).execute()
    (2,)

    """
    a = astensor(a)
    op = TensorGetShape(ndim=a.ndim)
    return op(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .cp import CpRuntimeOptimizer
from .ne import NeRuntimeOptimizer
from .dataframe import DataFrameRuntimeOptimizer


class RuntimeOptimizer:
    engine_dic = {
        'numpy': None,  # DO NOT optimize
        'numexpr': NeRuntimeOptimizer,
        'cupy': CpRuntimeOptimizer,
        'dataframe': DataFrameRuntimeOptimizer,
    }

    def __init__(self, graph, engine=None):
        # conver to DAG first if graph is not a DAG
        self._dag = graph.to_dag()
        self._graph = graph
        if engine is None:
            # add default optimizers
            engine = ['numexpr', 'cupy', 'dataframe']
        elif not isinstance(engine, (list, tuple)):
            engine = [engine]
        self._engine = engine

    def optimize(self, keys=None, check_availablility=True):
        self._dag.decompose()
        for e in self._engine:
            optimizer_cls = self.engine_dic[e]
            if optimizer_cls is None:
                continue
            if check_availablility and not optimizer_cls.is_available():
                continue
            optimizer = optimizer_cls(self._dag)
            optimizer.optimize(keys=keys)
        # copy back to the original graph,
        # only when the original graph is not a DAG
        self._dag.copyto(self._graph)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...operands import OperandStage
from ...serialize import BoolField
from ...utils import get_shuffle_input_keys_idxes, lazy_import
from ..operands import OutputType
from ..utils import parse_index, hash_dataframe_on, gen_unknown_index_value, standardize_range_index
from ._duplicate import DuplicateOperand, validate_subset

cudf = lazy_import('cudf', globals=globals())


class DataFrameDropDuplicates(DuplicateOperand):
    _op_type_ = opcodes.DROP_DUPLICATES

    _ignore_index = BoolField('ignore_index')

    def __init__(self, subset=None, keep=None, ignore_index=None,
                 output_types=None, method=None, subset_chunk=None,
                 shuffle_size=None, shuffle_phase=None, shuffle_key=None, **kw):
        super().__init__(_subset=subset, _keep=keep, _ignore_index=ignore_index,
                         _output_types=output_types, _method=method,
                         _subset_chunk=subset_chunk,
                         _shuffle_size=shuffle_size,
                         _shuffle_phase=shuffle_phase,
                         _shuffle_key=shuffle_key, **kw)

    @property
    def ignore_index(self):
        return self._ignore_index

    @classmethod
    def _get_shape(cls, input_shape):
        return (np.nan,) + input_shape[1:]

    @classmethod
    def _gen_tileable_params(cls, op: "DataFrameDropDuplicates", input_params):
        params = input_params.copy()
        if op.ignore_index:
            params['index_value'] = parse_index(pd.RangeIndex(-1))
        else:
            params['index_value'] = gen_unknown_index_value(
                input_params['index_value'], op.keep, op.subset, type(op).__name__)
        params['shape'] = cls._get_shape(input_params['shape'])
        return params

    def __call__(self, inp, inplace=False):
        self._output_types = inp.op.output_types
        params = self._gen_tileable_params(self, inp.params)

        ret = self.new_tileable([inp], kws=[params])
        if inplace:
            inp.data = ret.data
        return ret

    @classmethod
    def _gen_chunk_params(cls, op: "DataFrameDropDuplicates", input_chunk):
        input_params = input_chunk.params
        inp = op.inputs[0]
        chunk_params = input_params.copy()
        chunk_params['index'] = input_chunk.index[:1] + (0,) * (inp.ndim - 1)
        chunk_params['shape'] = cls._get_shape(input_params['shape'])
        chunk_params['index_value'] = gen_unknown_index_value(
            input_params['index_value'], input_chunk)
        if inp.ndim == 2:
            chunk_params['columns_value'] = inp.columns_value
            chunk_params['dtypes'] = inp.dtypes
        else:
            chunk_params['name'] = inp.name
            chunk_params['dtype'] = inp.dtype
        return chunk_params

    @classmethod
    def _tile_shuffle(cls, op: "DataFrameDropDuplicates", inp):
        tiled = super()._tile_shuffle(op, inp)[0]
        put_back_chunks = tiled.chunks
        if op.ignore_index:
            put_back_chunks = standardize_range_index(put_back_chunks)
        new_op = op.copy()
        params = tiled.params
        params['nsplits'] = tiled.nsplits
        params['chunks'] = put_back_chunks
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def _execute_chunk(cls, ctx, op):
        inp = ctx[op.input.key]
        ctx[op.outputs[0].key] = cls._drop_duplicates(inp, op)

    @classmethod
    def _execute_subset_tree_post(cls, ctx, op):
        inp = ctx[op.input.key]
        out = op.outputs[0]
        idx = op.outputs[0].index[0]
        subset = ctx[op.subset_chunk.key]
        selected = subset[subset['_chunk_index_'] == idx]['_i_']
        ret = inp.iloc[selected]
        if op.ignore_index:
            prev_size = (subset['_chunk_index_'] < out.index[0]).sum()
            ret.index = pd.RangeIndex(prev_size, prev_size + len(ret))
        ctx[op.outputs[0].key] = ret

    @classmethod
    def _execute_shuffle_map(cls, ctx, op):
        out = op.outputs[0]
        shuffle_size = op.shuffle_size
        subset = op.subset

        inp = ctx[op.input.key]
        dropped = cls._drop_duplicates(inp, op)
        if dropped.ndim == 1:
            dropped = dropped.to_frame()
            subset = dropped.columns.tolist()
        else:
            if subset is None:
                subset = dropped.columns.tolist()
        dropped['_chunk_index_'] = out.index[0]
        dropped['_i_'] = np.arange(dropped.shape[0])
        hashed = hash_dataframe_on(dropped, subset, shuffle_size)
        for i, data in enumerate(hashed):
            ctx[(out.key, str(i))] = dropped.iloc[data]

    @classmethod
    def _execute_shuffle_reduce(cls, ctx, op):
        out = op.outputs[0]

        input_keys, _ = get_shuffle_input_keys_idxes(op.inputs[0])
        inputs = [ctx[(inp_key, str(out.index[0]))] for inp_key in input_keys]
        xdf = cls._get_xdf(inputs[0])
        inp = xdf.concat(inputs)
        dropped = cls._drop_duplicates(inp, op,
                                       subset=[c for c in inp.columns
                                               if c not in ('_chunk_index_', '_i_')],
                                       keep=op.keep, ignore_index=op.ignore_index)
        for i in range(op.shuffle_size):
            filtered = dropped[dropped['_chunk_index_'] == i]
            del filtered['_chunk_index_']
            ctx[(out.key, str(i))] = filtered

    @classmethod
    def _execute_shuffle_put_back(cls, ctx, op):
        out = op.outputs[0]

        input_keys, _ = get_shuffle_input_keys_idxes(op.inputs[0])
        inputs = [ctx[(inp_key, str(out.index[0]))] for inp_key in input_keys]
        xdf = cls._get_xdf(inputs[0])
        inp = xdf.concat(inputs)
        inp.sort_values('_i_', inplace=True)
        del inp['_i_']

        if out.op.output_types[0] == OutputType.index:
            assert inp.shape[1] == 1
            ret = xdf.Index(inp.iloc[:, 0])
        elif out.op.output_types[0] == OutputType.series:
            assert inp.shape[1] == 1
            ret = inp.iloc[:, 0]
        else:
            ret = inp

        if op.ignore_index:
            ret.reset_index(drop=True, inplace=True)
        ctx[op.outputs[0].key] = ret

    @classmethod
    def execute(cls, ctx, op):
        if op.method is None:
            # one chunk
            cls._execute_chunk(ctx, op)
        elif op.method == 'tree':
            # tree
            cls._execute_chunk(ctx, op)
        elif op.method == 'subset_tree':
            # subset tree
            if op.stage == OperandStage.map:
                cls._execute_subset_tree_map(ctx, op)
            elif op.stage == OperandStage.combine:
                cls._execute_subset_tree_combine(ctx, op)
            elif op.stage == OperandStage.agg:
                cls._execute_subset_tree_agg(ctx, op)
            else:
                # post
                cls._execute_subset_tree_post(ctx, op)
        else:
            assert op.method == 'shuffle'
            if op.stage == OperandStage.map:
                cls._execute_shuffle_map(ctx, op)
            elif op.shuffle_phase == 'drop_duplicates':
                cls._execute_shuffle_reduce(ctx, op)
            else:
                assert op.shuffle_phase == 'put_back'
                cls._execute_shuffle_put_back(ctx, op)


def df_drop_duplicates(df, subset=None, keep='first',
                       inplace=False, ignore_index=False, method='auto'):
    """
    Return DataFrame with duplicate rows removed.

    Considering certain columns is optional. Indexes, including time indexes
    are ignored.

    Parameters
    ----------
    subset : column label or sequence of labels, optional
        Only consider certain columns for identifying duplicates, by
        default use all of the columns.
    keep : {'first', 'last', False}, default 'first'
        Determines which duplicates (if any) to keep.
        - ``first`` : Drop duplicates except for the first occurrence.
        - ``last`` : Drop duplicates except for the last occurrence.
        - False : Drop all duplicates.
    inplace : bool, default False
        Whether to drop duplicates in place or to return a copy.
    ignore_index : bool, default False
        If True, the resulting axis will be labeled 0, 1, …, n - 1.

    Returns
    -------
    DataFrame
        DataFrame with duplicates removed or None if ``inplace=True``.
    """
    if method not in ('auto', 'tree', 'subset_tree', 'shuffle', None):
        raise ValueError("method could only be one of "
                         "'auto', 'tree', 'subset_tree', 'shuffle' or None")
    subset = validate_subset(df, subset)
    op = DataFrameDropDuplicates(subset=subset, keep=keep,
                                 ignore_index=ignore_index,
                                 method=method)
    return op(df, inplace=inplace)


def series_drop_duplicates(series, keep='first', inplace=False, method='auto'):
    """
    Return Series with duplicate values removed.

    Parameters
    ----------
    keep : {'first', 'last', ``False``}, default 'first'
        Method to handle dropping duplicates:

        - 'first' : Drop duplicates except for the first occurrence.
        - 'last' : Drop duplicates except for the last occurrence.
        - ``False`` : Drop all duplicates.

    inplace : bool, default ``False``
        If ``True``, performs operation inplace and returns None.

    Returns
    -------
    Series
        Series with duplicates dropped.

    See Also
    --------
    Index.drop_duplicates : Equivalent method on Index.
    DataFrame.drop_duplicates : Equivalent method on DataFrame.
    Series.duplicated : Related method on Series, indicating duplicate
        Series values.

    Examples
    --------
    Generate a Series with duplicated entries.

    >>> import mars.dataframe as md
    >>> s = md.Series(['lama', 'cow', 'lama', 'beetle', 'lama', 'hippo'],
    ...               name='animal')
    >>> s.execute()
    0      lama
    1       cow
    2      lama
    3    beetle
    4      lama
    5     hippo
    Name: animal, dtype: object

    With the 'keep' parameter, the selection behaviour of duplicated values
    can be changed. The value 'first' keeps the first occurrence for each
    set of duplicated entries. The default value of keep is 'first'.

    >>> s.drop_duplicates().execute()
    0      lama
    1       cow
    3    beetle
    5     hippo
    Name: animal, dtype: object

    The value 'last' for parameter 'keep' keeps the last occurrence for
    each set of duplicated entries.

    >>> s.drop_duplicates(keep='last').execute()
    1       cow
    3    beetle
    4      lama
    5     hippo
    Name: animal, dtype: object

    The value ``False`` for parameter 'keep' discards all sets of
    duplicated entries. Setting the value of 'inplace' to ``True`` performs
    the operation inplace and returns ``None``.

    >>> s.drop_duplicates(keep=False, inplace=True)
    >>> s.execute()
    1       cow
    3    beetle
    5     hippo
    Name: animal, dtype: object
    """
    if method not in ('auto', 'tree', 'shuffle', None):
        raise ValueError("method could only be one of "
                         "'auto', 'tree', 'shuffle' or None")
    op = DataFrameDropDuplicates(keep=keep, method=method)
    return op(series, inplace=inplace)


def index_drop_duplicates(index, keep='first', method='auto'):
    """
    Return Index with duplicate values removed.

    Parameters
    ----------
    keep : {'first', 'last', ``False``}, default 'first'
        - 'first' : Drop duplicates except for the first occurrence.
        - 'last' : Drop duplicates except for the last occurrence.
        - ``False`` : Drop all duplicates.

    Returns
    -------
    deduplicated : Index

    See Also
    --------
    Series.drop_duplicates : Equivalent method on Series.
    DataFrame.drop_duplicates : Equivalent method on DataFrame.
    Index.duplicated : Related method on Index, indicating duplicate
        Index values.

    Examples
    --------
    Generate an pandas.Index with duplicate values.

    >>> import mars.dataframe as md

    >>> idx = md.Index(['lama', 'cow', 'lama', 'beetle', 'lama', 'hippo'])

    The `keep` parameter controls  which duplicate values are removed.
    The value 'first' keeps the first occurrence for each
    set of duplicated entries. The default value of keep is 'first'.

    >>> idx.drop_duplicates(keep='first').execute()
    Index(['lama', 'cow', 'beetle', 'hippo'], dtype='object')

    The value 'last' keeps the last occurrence for each set of duplicated
    entries.

    >>> idx.drop_duplicates(keep='last').execute()
    Index(['cow', 'beetle', 'lama', 'hippo'], dtype='object')

    The value ``False`` discards all sets of duplicated entries.

    >>> idx.drop_duplicates(keep=False).execute()
    Index(['cow', 'beetle', 'hippo'], dtype='object')
    """
    if method not in ('auto', 'tree', 'shuffle', None):
        raise ValueError("method could only be one of "
                         "'auto', 'tree', 'shuffle' or None")
    op = DataFrameDropDuplicates(keep=keep, method=method)
    return op(index)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import array
from enum import Enum

from ..cluster_info import ClusterInfoActor, HasClusterInfoActor
from ..utils import classproperty
from ..promise import PromiseActor


class GraphState(Enum):
    UNSCHEDULED = 'unscheduled'
    PREPARING = 'preparing'
    RUNNING = 'running'
    SUCCEEDED = 'succeeded'
    FAILED = 'failed'
    CANCELLING = 'cancelling'
    CANCELLED = 'cancelled'

    @classproperty
    def TERMINATED_STATES(self):
        """
        States on which the graph has already terminated
        """
        return self.SUCCEEDED, self.FAILED


class SchedulerClusterInfoActor(ClusterInfoActor):
    @classmethod
    def default_uid(cls):
        return f's:h1:{cls.__name__}'


class SchedulerHasClusterInfoActor(HasClusterInfoActor):
    cluster_info_uid = SchedulerClusterInfoActor.default_uid()


class SchedulerActor(SchedulerHasClusterInfoActor, PromiseActor):
    @classmethod
    def default_uid(cls):
        return f's:h1:{cls.__name__}'

    @property
    def chunk_meta(self):
        try:
            return self._chunk_meta_client
        except AttributeError:
            from .chunkmeta import ChunkMetaClient
            self._chunk_meta_client = ChunkMetaClient(self.ctx, self._cluster_info_ref)
            return self._chunk_meta_client


class CombinedFutureWaiter(object):
    def __init__(self, futures):
        self._futures = futures

    def result(self):
        return [f.result() for f in self._futures]


def array_to_bytes(typecode, initializer):
    return array.array(typecode, initializer).tobytes()



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator

from ... import opcodes as OperandDef
from ...utils import classproperty
from .core import DataFrameBinopUfunc


class DataFrameXor(DataFrameBinopUfunc):
    _op_type_ = OperandDef.XOR

    _bit_func_name = '__xor__'
    _bit_rfunc_name = '__rxor__'

    @classproperty
    def _operator(self):
        return operator.xor

    @classproperty
    def tensor_op_type(self):
        from ...tensor.arithmetic import TensorBitxor
        return TensorBitxor


def bitxor(df, other, axis='columns', level=None, fill_value=None):
    op = DataFrameXor(axis=axis, level=level, fill_value=fill_value, lhs=df, rhs=other)
    return op(df, other)


def rbitxor(df, other, axis='columns', level=None, fill_value=None):
    op = DataFrameXor(axis=axis, level=level, fill_value=fill_value, lhs=other, rhs=df)
    return op.rcall(df, other)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .array import SparseArray, SparseNDArray
from .core import get_array_module, get_sparse_module, naked, issparse, np, is_cupy


class SparseVector(SparseArray):
    __slots__ = 'spmatrix',

    def __init__(self, spvector, shape=()):
        if shape and len(shape) != 1:
            raise ValueError('Only accept 1-d array')
        if isinstance(spvector, SparseVector):
            self.spmatrix = spvector.spmatrix
        else:
            spvector = spvector.reshape(1, shape[0])
            self.spmatrix = spvector.tocsr()

    @property
    def shape(self):
        return self.spmatrix.shape[1],

    def transpose(self, axes=None):
        assert axes is None or tuple(axes) == (0,)
        return self

    @property
    def T(self):
        return self

    def __truediv__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        x = self.spmatrix / other
        if issparse(x):
            return SparseNDArray(x, shape=self.shape)
        if x.shape != self.shape:
            x = np.asarray(x).reshape(self.shape)
        return get_array_module(x).asarray(x)

    def __rtruediv__(self, other):
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented
        try:
            x = other / self.spmatrix
        except TypeError:
            x = other / self.spmatrix.toarray()
        if issparse(x):
            return SparseNDArray(x, shape=self.shape)
        if x.shape != self.shape:
            x = np.asarray(x).reshape(self.shape)
        return get_array_module(x).asarray(x)

    def dot(self, other, sparse=True):
        other_shape = other.shape
        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if not sparse:
            a = self.toarray()
            if issparse(other):
                other = other.toarray().reshape(other_shape)

            x = a.dot(other)
        else:
            if len(other_shape) == 1:
                x = self.spmatrix.dot(other.T)
            else:
                x = self.spmatrix.dot(other)
        if issparse(x):
            if x.shape == (1, 1):
                # return scalar
                return x.toarray()[0, 0]
            shape = (x.shape[1],)
            return SparseNDArray(x, shape=shape)
        return get_array_module(x).asarray(x)

    def concatenate(self, other, axis=0):
        if other.ndim != 1:
            raise ValueError('all the input arrays must have same number of dimensions')

        try:
            other = naked(other)
        except TypeError:
            return NotImplemented

        if issparse(other):
            xps = get_sparse_module(self.spmatrix)
            if axis != 0:
                raise ValueError('axis can only be 0')
            other = other.reshape(1, other.shape[0]) if other.shape[0] != 1 else other
            x = xps.hstack((self.spmatrix.reshape(1, self.shape[0]), other))
        else:
            xp = get_array_module(self.spmatrix)
            x = xp.concatenate((self.spmatrix.toarray().reshape(self.shape), other), axis=axis)

        if issparse(x):
            return SparseNDArray(x, shape=(x.shape[1],))
        return get_array_module(x).asarray(x)

    def _reduction(self, method_name, axis=None, dtype=None, keepdims=None, todense=False, **kw):
        if not todense:
            assert keepdims is None or keepdims is False

        if isinstance(axis, tuple):
            assert axis == (0, )
            axis = None

        if todense:
            x = self.spmatrix.toarray()
            x = getattr(get_array_module(x), method_name)(x, axis=axis, **kw)
        else:
            x = getattr(self.spmatrix, method_name)(axis=axis, **kw)

        m = get_array_module(x)
        return m.array([x])[0]

    def __setitem__(self, key, value):
        if is_cupy(self.spmatrix):
            return NotImplemented
        else:
            x = self.spmatrix.tolil()
            key = (0,) + (key, )
            x[key] = value
            x = x.tocsr()
        self.spmatrix = x



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging
import os
import pickle  # nosec  # pylint: disable=import_pickle
import tempfile
import uuid

import numpy as np

from .... import opcodes
from .... import tensor as mt
from ....context import get_context, RunningMode
from ....lib.filesystem import get_fs, LocalFileSystem
from ....operands import OutputType, OperandStage
from ....serialize import StringField, Int32Field, Int64Field, \
    DictField, BytesField, TupleField, DataTypeField
from ....tiles import TilesError
from ....utils import check_chunks_unknown_shape, Timer
from ...operands import LearnOperand, LearnOperandMixin
from ..core import proxima, get_proxima_type, validate_tensor, \
    available_numpy_dtypes, rechunk_tensor, build_mmap_chunks

logger = logging.getLogger(__name__)

DEFAULT_INDEX_SIZE = 5 * 10 ** 6


class ProximaBuilder(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.PROXIMA_SIMPLE_BUILDER

    _distance_metric = StringField('distance_metric')
    _dimension = Int32Field('dimension')
    _column_number = Int64Field('column_number')
    _index_path = StringField('index_path')
    _index_builder = StringField('index_builder')
    _index_builder_params = DictField('index_builder_params')
    _index_converter = StringField('index_converter')
    _index_converter_params = DictField('index_converter_params')
    _topk = Int32Field('topk')
    _storage_options = BytesField('storage_options',
                                  on_serialize=pickle.dumps,
                                  on_deserialize=pickle.loads)

    # only for chunk
    _array_shape = TupleField('array_shape')
    _array_dtype = DataTypeField('array_dtype')
    _offset = Int64Field('offset')

    def __init__(self, distance_metric=None, index_path=None,
                 dimension=None, column_number=None,
                 index_builder=None, index_builder_params=None,
                 index_converter=None, index_converter_params=None,
                 array_shape=None, array_dtype=None, offset=None,
                 topk=None, storage_options=None, output_types=None, stage=None, **kw):
        super().__init__(_distance_metric=distance_metric, _index_path=index_path,
                         _dimension=dimension, _column_number=column_number, _index_builder=index_builder,
                         _index_builder_params=index_builder_params,
                         _array_shape=array_shape, _array_dtype=array_dtype, _offset=offset,
                         _index_converter=index_converter, _index_converter_params=index_converter_params,
                         _topk=topk, _storage_options=storage_options,
                         _output_types=output_types, _stage=stage, **kw)
        if self._output_types is None:
            self._output_types = [OutputType.object]

    @property
    def distance_metric(self):
        return self._distance_metric

    @property
    def column_number(self):
        return self._column_number

    @property
    def index_path(self):
        return self._index_path

    @property
    def dimension(self):
        return self._dimension

    @property
    def index_builder(self):
        return self._index_builder

    @property
    def index_builder_params(self):
        return self._index_builder_params

    @property
    def index_converter(self):
        return self._index_converter

    @property
    def index_converter_params(self):
        return self._index_converter_params

    @property
    def topk(self):
        return self._topk

    @property
    def storage_options(self):
        return self._storage_options

    @property
    def array_shape(self):
        return self._array_shape

    @property
    def array_dtype(self):
        return self._array_dtype

    @property
    def offset(self):
        return self._offset

    def __call__(self, tensor):
        return self.new_tileable([tensor])

    @classmethod
    def _get_atleast_topk_nsplit(cls, nsplit, topk):
        new_nsplit = []
        i = 0
        while i < len(nsplit):
            cur = nsplit[i]
            i += 1
            if cur >= topk:
                new_nsplit.append(cur)
            else:
                while i < len(nsplit):
                    cur += nsplit[i]
                    i += 1
                    if cur >= topk:
                        break
                if cur < topk and len(new_nsplit) > 0:
                    new_nsplit[-1] += cur
                elif cur >= topk:
                    new_nsplit.append(cur)
        new_nsplit = tuple(new_nsplit)
        assert sum(new_nsplit) == sum(nsplit), f'sum of nsplit not equal, ' \
                                               f'old: {nsplit}, new: {new_nsplit}'

        return new_nsplit

    @classmethod
    def tile(cls, op):
        tensor = op.inputs[0]
        out = op.outputs[0]
        index_path = op.index_path
        ctx = get_context()
        fs = None
        if index_path is not None:
            fs = get_fs(index_path, op.storage_options)

        # check index_path for distributed
        if getattr(ctx, 'running_mode', None) == RunningMode.distributed:
            if index_path is not None:
                if isinstance(fs, LocalFileSystem):
                    raise ValueError('`index_path` cannot be local file dir '
                                     'for distributed index building')

        if index_path is not None:
            # check if the index path is empty
            try:
                files = [f for f in fs.ls(index_path) if 'proxima_' in f]
                if files:
                    raise ValueError(f'Directory {index_path} contains built proxima index, '
                                     f'clean them to perform new index building')
            except FileNotFoundError:
                # if not exist, create directory
                fs.mkdir(index_path)

        # make sure all inputs have known chunk sizes
        check_chunks_unknown_shape(op.inputs, TilesError)

        if op.column_number:
            index_chunk_size = op.inputs[0].shape[0] // op.column_number
        else:
            worker_num = len(ctx.get_worker_addresses() or [])
            if worker_num > 0:
                index_chunk_size = max(op.inputs[0].shape[0] // worker_num, DEFAULT_INDEX_SIZE)
            else:
                index_chunk_size = DEFAULT_INDEX_SIZE

        if op.topk is not None:
            index_chunk_size = cls._get_atleast_topk_nsplit(index_chunk_size, op.topk)

        # build chunks for writing tensors to mmap files.
        worker_iter = iter(itertools.cycle(ctx.get_worker_addresses() or [None]))
        chunk_groups = rechunk_tensor(tensor, index_chunk_size)
        out_chunks = []
        offsets = []
        offset = 0
        for chunk_group in chunk_groups:
            offsets.append(offset)
            file_prefix = f'proxima-build-{str(uuid.uuid4())}'
            out_chunks.append(build_mmap_chunks(chunk_group, next(worker_iter),
                                                file_prefix=file_prefix))
            offset += sum(c.shape[0] for c in chunk_group)

        final_out_chunks = []
        for j, chunks in enumerate(out_chunks):
            chunk_op = op.copy().reset_key()
            chunk_op._stage = OperandStage.map
            chunk_op._expect_worker = chunks[0].op.expect_worker
            chunk_op._array_shape = chunks[0].op.total_shape
            chunk_op._array_dtype = chunks[0].dtype
            chunk_op._offset = offsets[j]
            out_chunk = chunk_op.new_chunk(chunks, index=(j,))
            final_out_chunks.append(out_chunk)

        logger.warning(f"index chunks count: {len(final_out_chunks)} ")

        params = out.params
        params['chunks'] = final_out_chunks
        params['nsplits'] = ((1,) * len(final_out_chunks),)
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def _execute_map(cls, ctx, op: "ProximaBuilder"):
        mmap_path = ctx[op.inputs[0].key]
        out = op.outputs[0]

        data = np.memmap(mmap_path, dtype=op.array_dtype, mode='r',
                         shape=op.array_shape)

        proxima_type = get_proxima_type(op.array_dtype)
        offset = op.offset

        # holder
        with Timer() as timer:
            holder = proxima.IndexHolder(type=proxima_type,
                                         dimension=op.dimension, shallow=True)
            holder.mount(data, key_base=offset)

        logger.warning(f'Holder({op.key}) costs {timer.duration} seconds')

        # converter
        meta = proxima.IndexMeta(proxima_type, dimension=op.dimension,
                                 measure_name=op.distance_metric)
        if op.index_converter is not None:
            with Timer() as timer:
                converter = proxima.IndexConverter(name=op.index_converter,
                                                   meta=meta, params=op.index_converter_params)
                converter.train_and_transform(holder)
                holder = converter.result()
                meta = converter.meta()

            logger.warning(f'Converter({op.key}) costs {timer.duration} seconds')

        # builder
        with Timer() as timer:
            builder = proxima.IndexBuilder(name=op.index_builder,
                                           meta=meta,
                                           params=op.index_builder_params)
            builder = builder.train_and_build(holder)

        logger.warning(f'Builder({op.key}) costs {timer.duration} seconds')

        # remove mmap file
        os.remove(mmap_path)

        # dumper
        with Timer() as timer:
            path = tempfile.mkstemp(prefix='proxima-', suffix='.index')[1]
            dumper = proxima.IndexDumper(name="FileDumper", path=path)
            builder.dump(dumper)
            dumper.close()

        logger.warning(f'Dumper({op.key}) costs {timer.duration} seconds')

        if op.index_path is None:
            ctx[out.key] = path
        else:
            # write to external file
            with Timer() as timer:
                fs = get_fs(op.index_path, op.storage_options)
                filename = f'proxima_{out.index[0]}_index'
                out_path = f'{op.index_path.rstrip("/")}/{filename}'

                def write_index():
                    with fs.open(out_path, 'wb') as out_f:
                        with open(path, 'rb') as in_f:
                            # 128M
                            chunk_bytes = 128 * 1024 ** 2
                            while True:
                                data = in_f.read(chunk_bytes)
                                if data:
                                    out_f.write(data)
                                else:
                                    break

                # retry 3 times
                for _ in range(3):
                    try:
                        write_index()
                        break
                    except:  # noqa: E722  # nosec  # pylint: disable=bare-except
                        fs.delete(out_path)
                        continue

            logger.warning(f'WritingToVolume({op.key}), out path: {out_path}, '
                           f'size {os.path.getsize(path)}, '
                           f'costs {timer.duration} seconds '
                           f'speed {round(os.path.getsize(path) / (1024 ** 2) / timer.duration, 2)} MB/s')

            ctx[out.key] = filename

    @classmethod
    def _execute_agg(cls, ctx, op: "ProximaBuilder"):
        paths = [ctx[inp.key] for inp in op.inputs]
        ctx[op.outputs[0].key] = paths

    @classmethod
    def execute(cls, ctx, op: "ProximaBuilder"):
        if op.stage != OperandStage.agg:
            return cls._execute_map(ctx, op)
        else:
            return cls._execute_agg(ctx, op)

    @classmethod
    def concat_tileable_chunks(cls, tileable):
        assert not tileable.is_coarse()

        op = cls(stage=OperandStage.agg)
        chunk = cls(stage=OperandStage.agg).new_chunk(tileable.chunks)
        return op.new_tileable([tileable], chunks=[chunk], nsplits=((1,),))


def build_index(tensor, dimension=None, index_path=None, column_number=None,
                need_shuffle=False, distance_metric='SquaredEuclidean',
                index_builder='SsgBuilder', index_builder_params=None,
                index_converter=None, index_converter_params=None,
                topk=None, storage_options=None,
                run=True, session=None, run_kwargs=None):
    tensor = validate_tensor(tensor)
    if tensor.dtype not in available_numpy_dtypes:
        raise ValueError(f'Dtype to build index should be one of {available_numpy_dtypes}, '
                         f'got {tensor.dtype}')

    if dimension is None:
        dimension = tensor.shape[1]
    if index_builder_params is None:
        index_builder_params = {}
    if index_converter_params is None:
        index_converter_params = {}

    if need_shuffle:
        tensor = mt.random.permutation(tensor)

    op = ProximaBuilder(distance_metric=distance_metric,
                        index_path=index_path, dimension=dimension,
                        column_number=column_number,
                        index_builder=index_builder,
                        index_builder_params=index_builder_params,
                        index_converter=index_converter,
                        index_converter_params=index_converter_params,
                        topk=topk, storage_options=storage_options)
    result = op(tensor)
    if run:
        return result.execute(session=session, **(run_kwargs or dict()))
    else:
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorUnaryOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode='unary')
class TensorSign(TensorUnaryOp):
    _op_type_ = OperandDef.SIGN
    _func_name = 'sign'


@infer_dtype(np.sign)
def sign(x, out=None, where=None, **kwargs):
    r"""
    Returns an element-wise indication of the sign of a number.

    The `sign` function returns ``-1 if x < 0, 0 if x==0, 1 if x > 0``.  nan
    is returned for nan inputs.

    For complex inputs, the `sign` function returns
    ``sign(x.real) + 0j if x.real != 0 else sign(x.imag) + 0j``.

    complex(nan, 0) is returned for complex nan inputs.

    Parameters
    ----------
    x : array_like
      Input values.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : Tensor
      The sign of `x`.

    Notes
    -----
    There is more than one definition of sign in common use for complex
    numbers.  The definition used here is equivalent to :math:`x/\sqrt{x*x}`
    which is different from a common alternative, :math:`x/|x|`.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.sign([-5., 4.5]).execute()
    array([-1.,  1.])
    >>> mt.sign(0).execute()
    0
    >>> mt.sign(5-2j).execute()
    (1+0j)
    """
    op = TensorSign(**kwargs)
    return op(x, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorBinOp
from .utils import arithmetic_operand


@arithmetic_operand
class TensorDivide(TensorBinOp):
    _op_type_ = OperandDef.DIV
    _func_name = 'divide'

    @classmethod
    def _is_sparse(cls, x1, x2):
        if not np.isscalar(x1) and not np.isscalar(x2):
            return False
        if hasattr(x1, 'issparse') and x1.issparse():
            if x2 != 0:
                return True
            else:
                raise ZeroDivisionError('float division by zero')


@infer_dtype(np.divide)
def divide(x1, x2, out=None, where=None, **kwargs):
    """
    Divide arguments element-wise.

    Parameters
    ----------
    x1 : array_like
        Dividend tensor.
    x2 : array_like
        Divisor tensor.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated array is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    out : Tensor
        The quotient `x1/x2`, element-wise. Returns a scalar if both `x1` and `x2` are scalars.

    Notes
    -----
    Equivalent to `x1` / `x2` in terms of array-broadcasting.

    Behavior on division by zero can be changed using `seterr`.

    In Python 2, when both `x1` and `x2` are of an integer type, `divide` will behave like `floor_divide`.
    In Python 3, it behaves like `true_divide`.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.divide(2.0, 4.0).execute()
    0.5
    >>> x1 = mt.arange(9.0).reshape((3, 3))
    >>> x2 = mt.arange(3.0)
    >>> mt.divide(x1, x2).execute()
    array([[ NaN,  1. ,  1. ],
           [ Inf,  4. ,  2.5],
           [ Inf,  7. ,  4. ]])
    Note the behavior with integer types (Python 2 only):
    >>> mt.divide(2, 4).execute()
    0
    >>> mt.divide(2, 4.).execute()
    0.5
    Division by zero always yields zero in integer arithmetic (again, Python 2 only),
    and does not raise an exception or a warning:
    >>> mt.divide(mt.array([0, 1], dtype=int), mt.array([0, 0], dtype=int)).execute()
    array([0, 0])
    Division by zero can, however, be caught using seterr:
    >>> old_err_state = mt.seterr(divide='raise')
    >>> mt.divide(1, 0).execute()
    Traceback (most recent call last):
    ...
    FloatingPointError: divide by zero encountered in divide
    >>> ignored_states = mt.seterr(**old_err_state)
    >>> mt.divide(1, 0).execute()
    0
    """
    op = TensorDivide(**kwargs)
    return op(x1, x2, out=out, where=where)


@infer_dtype(np.divide, reverse=True)
def rdivide(x1, x2, **kwargs):
    op = TensorDivide(**kwargs)
    return op.rcall(x1, x2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import AnyField, TupleField
from .core import TensorReduction, TensorArgReductionMixin


class TensorNanArgmax(TensorReduction, TensorArgReductionMixin):
    _op_type_ = OperandDef.NANARGMAX
    _func_name = 'nanargmax'
    _agg_func_name = 'nanmax'

    _offset = AnyField('offset')
    _total_shape = TupleField('total_shape')

    def __init__(self, axis=None, dtype=np.dtype(int), combine_size=None,
                 offset=None, total_shape=None, stage=None, **kw):
        stage = self._rewrite_stage(stage)
        super().__init__(_axis=axis, _dtype=dtype, _combine_size=combine_size,
                         _offset=offset, _total_shape=total_shape, _stage=stage, **kw)

    @property
    def offset(self):
        return getattr(self, '_offset', None)

    @property
    def total_shape(self):
        return getattr(self, '_total_shape', None)


def nanargmax(a, axis=None, out=None, combine_size=None):
    """
    Return the indices of the maximum values in the specified axis ignoring
    NaNs. For all-NaN slices ``ValueError`` is raised. Warning: the
    results cannot be trusted if a slice contains only NaNs and -Infs.


    Parameters
    ----------
    a : array_like
        Input data.
    axis : int, optional
        Axis along which to operate.  By default flattened input is used.
    out : Tensor, optional
        Alternate output tensor in which to place the result.  The default
        is ``None``; if provided, it must have the same shape as the
        expected output, but the type will be cast if necessary.
        See `doc.ufuncs` for details.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    index_array : Tensor
        An tensor of indices or a single index value.

    See Also
    --------
    argmax, nanargmin

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array([[mt.nan, 4], [2, 3]])
    >>> mt.argmax(a).execute()
    0
    >>> mt.nanargmax(a).execute()
    1
    >>> mt.nanargmax(a, axis=0).execute()
    array([1, 0])
    >>> mt.nanargmax(a, axis=1).execute()
    array([1, 1])

    """
    op = TensorNanArgmax(axis=axis, dtype=np.dtype(int), combine_size=combine_size)
    return op(a, out=out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

from ... import opcodes as OperandDef
from ...utils import lazy_import
from ...serialize import Int32Field
from ..operands import DataFrameOperandMixin, DataFrameOperand


cudf = lazy_import('cudf', globals=globals())


class ChunkStandardizeRangeIndex(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.STANDARDIZE_RANGE_INDEX

    _axis = Int32Field('axis')

    def __init__(self, pure_depends=None, axis=None, output_types=None, **kwargs):
        super().__init__(_pure_depends=pure_depends, _axis=axis, _output_types=output_types, **kwargs)

    @property
    def axis(self):
        return self._axis

    @classmethod
    def execute(cls, ctx, op):
        xdf = cudf if op.gpu else pd
        in_data = ctx[op.inputs[-1].key].copy()
        input_keys = [c.key for c in op.inputs[:-1]]
        metas = ctx.get_chunk_metas(input_keys)
        index_start = sum([m.chunk_shape[op.axis] for m in metas])
        if op.axis == 0:
            in_data.index = xdf.RangeIndex(index_start, index_start + len(in_data))
        else:
            in_data.columns = xdf.RangeIndex(index_start, index_start + in_data.shape[1])
        ctx[op.outputs[0].key] = in_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

cimport numpy as np

from ..core cimport ActorRef


ctypedef np.int64_t INT64_t
ctypedef np.int32_t INT32_t
ctypedef np.int16_t INT16_t
ctypedef np.uint8_t BYTE_t


cpdef enum MessageType:
    create_actor = 0
    destroy_actor = 1
    has_actor = 2
    result = 3
    error = 4
    send_all = 5
    tell_all = 6
    send_chunk_start = 7
    send_chunk = 8
    send_chunk_end = 9
    tell_chunk_start = 10
    tell_chunk = 11
    tell_chunk_end = 12


cpdef INT32_t unpack_message_type_value(bytes binary)
cpdef object unpack_message_type(bytes binary)
cpdef bytes unpack_message_id(bytes binary)
cpdef bytes pack_actor_ref(ActorRef actor_ref)
cpdef void unpack_actor_ref(bytes binary, ActorRef actor_ref)
cpdef object unpack_send_message(bytes message)
cpdef object pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,
                               object write=*, INT32_t protocol=*)
cpdef object unpack_send_message(bytes message)
cpdef object pack_tell_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,
                               object write=*, INT32_t protocol=*)
cpdef object unpack_tell_message(bytes message)
cpdef tuple get_index(bytes binary, object calc_from_uid)
cpdef object pack_create_actor_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object actor_cls,
                                       tuple args, dict kw, object write=*, INT32_t protocol=*)
cpdef object unpack_create_actor_message(bytes binary)
cpdef object pack_destroy_actor_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref,
                                        object write=*, INT32_t protocol=*)
cpdef object unpack_destroy_actor_message(bytes binary)
cpdef object pack_has_actor_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref,
                                    object write=*, INT32_t protocol=*)
cpdef object unpack_has_actor_message(bytes binary)
cpdef object pack_result_message(bytes message_id, INT32_t from_index, INT32_t to_index, object result,
                                 object write=*, INT32_t protocol=*)
cpdef object unpack_result_message(bytes binary, object from_index=*, object to_index=*)
cpdef object pack_error_message(bytes message_id, INT32_t from_index, INT32_t to_index,
                                object error_type, object error, object tb,
                                object write=*, INT32_t protocol=*)
cpdef object unpack_error_message(bytes binary)
cpdef bytes read_remote_message(object read_func)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from .core import TensorRandomOperandMixin, TensorSimpleRandomData


class TensorRand(TensorSimpleRandomData, TensorRandomOperandMixin):
    _op_type_ = OperandDef.RAND_RAND
    _func_name = 'rand'

    def __init__(self, state=None, size=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super().__init__(_state=state, _size=size, _dtype=dtype, _gpu=gpu, **kw)

    def __call__(self, chunk_size=None):
        return self.new_tensor(None, None, raw_chunk_size=chunk_size)


def rand(random_state, *dn, **kw):
    """
    Random values in a given shape.

    Create a tensor of the given shape and populate it with
    random samples from a uniform distributionc
    over ``[0, 1)``.

    Parameters
    ----------
    d0, d1, ..., dn : int, optional
        The dimensions of the returned tensor, should all be positive.
        If no argument is given a single Python float is returned.

    Returns
    -------
    out : Tensor, shape ``(d0, d1, ..., dn)``
        Random values.

    See Also
    --------
    random

    Notes
    -----
    This is a convenience function. If you want an interface that
    takes a shape-tuple as the first argument, refer to
    mt.random.random_sample .

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.random.rand(3, 2).execute()
    array([[ 0.14022471,  0.96360618],  #random
           [ 0.37601032,  0.25528411],  #random
           [ 0.49313049,  0.94909878]]) #random
    """
    if len(dn) == 1 and isinstance(dn[0], (tuple, list)):
        raise TypeError("'tuple' object cannot be interpreted as an integer")
    if 'dtype' not in kw:
        kw['dtype'] = np.dtype('f8')
    chunk_size = kw.pop('chunk_size', None)
    op = TensorRand(state=random_state.to_numpy(), size=dn, **kw)

    for key in op.extra_params:
        if not key.startswith('_'):
            raise ValueError(f'rand got unexpected key arguments {key}')

    return op(chunk_size=chunk_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from ..arithmetic.add import TensorTreeAdd
from .core import TensorCumReduction, TensorCumReductionMixin


class TensorNanCumsum(TensorCumReduction, TensorCumReductionMixin):
    _op_type_ = OperandDef.NANCUMSUM
    _func_name = 'nancumsum'

    def __init__(self, axis=None, dtype=None, **kw):
        super().__init__(_axis=axis, _dtype=dtype, **kw)

    @staticmethod
    def _get_op_types():
        return TensorNanCumsum, TensorTreeAdd


def nancumsum(a, axis=None, dtype=None, out=None):
    """
    Return the cumulative sum of tensor elements over a given axis treating Not a
    Numbers (NaNs) as zero.  The cumulative sum does not change when NaNs are
    encountered and leading NaNs are replaced by zeros.

    Zeros are returned for slices that are all-NaN or empty.

    Parameters
    ----------
    a : array_like
        Input tensor.
    axis : int, optional
        Axis along which the cumulative sum is computed. The default
        (None) is to compute the cumsum over the flattened tensor.
    dtype : dtype, optional
        Type of the returned tensor and of the accumulator in which the
        elements are summed.  If `dtype` is not specified, it defaults
        to the dtype of `a`, unless `a` has an integer dtype with a
        precision less than that of the default platform integer.  In
        that case, the default platform integer is used.
    out : Tensor, optional
        Alternative output tensor in which to place the result. It must
        have the same shape and buffer length as the expected output
        but the type will be cast if necessary. See `doc.ufuncs`
        (Section "Output arguments") for more details.

    Returns
    -------
    nancumsum : Tensor.
        A new tensor holding the result is returned unless `out` is
        specified, in which it is returned. The result has the same
        size as `a`, and the same shape as `a` if `axis` is not None
        or `a` is a 1-d tensor.

    See Also
    --------
    numpy.cumsum : Cumulative sum across tensor propagating NaNs.
    isnan : Show which elements are NaN.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.nancumsum(1).execute()
    array([1])
    >>> mt.nancumsum([1]).execute()
    array([1])
    >>> mt.nancumsum([1, mt.nan]).execute()
    array([ 1.,  1.])
    >>> a = mt.array([[1, 2], [3, mt.nan]])
    >>> mt.nancumsum(a).execute()
    array([ 1.,  3.,  6.,  6.])
    >>> mt.nancumsum(a, axis=0).execute()
    array([[ 1.,  2.],
           [ 4.,  2.]])
    >>> mt.nancumsum(a, axis=1).execute()
    array([[ 1.,  3.],
           [ 3.,  3.]])

    """
    a = astensor(a)
    if dtype is None:
        dtype = np.nancumsum(np.empty((1,), dtype=a.dtype)).dtype
    op = TensorNanCumsum(axis=axis, dtype=dtype)
    return op(a, out=out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorUnaryOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode='unary')
class TensorLog10(TensorUnaryOp):
    _op_type_ = OperandDef.LOG10
    _func_name = 'log10'


@infer_dtype(np.log10)
def log10(x, out=None, where=None, **kwargs):
    """
    Return the base 10 logarithm of the input tensor, element-wise.

    Parameters
    ----------
    x : array_like
        Input values.
    out : Tensor, None, or tuple of tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : Tensor
        The logarithm to the base 10 of `x`, element-wise. NaNs are
        returned where x is negative.

    Notes
    -----
    Logarithm is a multivalued function: for each `x` there is an infinite
    number of `z` such that `10**z = x`. The convention is to return the
    `z` whose imaginary part lies in `[-pi, pi]`.

    For real-valued input data types, `log10` always returns real output.
    For each value that cannot be expressed as a real number or infinity,
    it yields ``nan`` and sets the `invalid` floating point error flag.

    For complex-valued input, `log10` is a complex analytical function that
    has a branch cut `[-inf, 0]` and is continuous from above on it.
    `log10` handles the floating-point negative zero as an infinitesimal
    negative number, conforming to the C99 standard.

    References
    ----------
    .. [1] M. Abramowitz and I.A. Stegun, "Handbook of Mathematical Functions",
           10th printing, 1964, pp. 67. http://www.math.sfu.ca/~cbm/aands/
    .. [2] Wikipedia, "Logarithm". http://en.wikipedia.org/wiki/Logarithm

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.log10([1e-15, -3.]).execute()
    array([-15.,  NaN])
    """
    op = TensorLog10(**kwargs)
    return op(x, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from .core import TensorRealFFTNMixin, validate_fftn, TensorRealFFTN


class TensorRFFTN(TensorRealFFTN, TensorRealFFTNMixin):
    _op_type_ = OperandDef.RFFTN

    def __init__(self, shape=None, axes=None, norm=None, dtype=None, **kw):
        super().__init__(_shape=shape, _axes=axes, _norm=norm, _dtype=dtype, **kw)


def rfftn(a, s=None, axes=None, norm=None):
    """
    Compute the N-dimensional discrete Fourier Transform for real input.

    This function computes the N-dimensional discrete Fourier Transform over
    any number of axes in an M-dimensional real tensor by means of the Fast
    Fourier Transform (FFT).  By default, all axes are transformed, with the
    real transform performed over the last axis, while the remaining
    transforms are complex.

    Parameters
    ----------
    a : array_like
        Input tensor, taken to be real.
    s : sequence of ints, optional
        Shape (length along each transformed axis) to use from the input.
        (``s[0]`` refers to axis 0, ``s[1]`` to axis 1, etc.).
        The final element of `s` corresponds to `n` for ``rfft(x, n)``, while
        for the remaining axes, it corresponds to `n` for ``fft(x, n)``.
        Along any axis, if the given shape is smaller than that of the input,
        the input is cropped.  If it is larger, the input is padded with zeros.
        if `s` is not given, the shape of the input along the axes specified
        by `axes` is used.
    axes : sequence of ints, optional
        Axes over which to compute the FFT.  If not given, the last ``len(s)``
        axes are used, or all axes if `s` is also not specified.
    norm : {None, "ortho"}, optional
        Normalization mode (see `mt.fft`). Default is None.

    Returns
    -------
    out : complex Tensor
        The truncated or zero-padded input, transformed along the axes
        indicated by `axes`, or by a combination of `s` and `a`,
        as explained in the parameters section above.
        The length of the last axis transformed will be ``s[-1]//2+1``,
        while the remaining transformed axes will have lengths according to
        `s`, or unchanged from the input.

    Raises
    ------
    ValueError
        If `s` and `axes` have different length.
    IndexError
        If an element of `axes` is larger than than the number of axes of `a`.

    See Also
    --------
    irfftn : The inverse of `rfftn`, i.e. the inverse of the n-dimensional FFT
         of real input.
    fft : The one-dimensional FFT, with definitions and conventions used.
    rfft : The one-dimensional FFT of real input.
    fftn : The n-dimensional FFT.
    rfft2 : The two-dimensional FFT of real input.

    Notes
    -----
    The transform for real input is performed over the last transformation
    axis, as by `rfft`, then the transform over the remaining axes is
    performed as by `fftn`.  The order of the output is as for `rfft` for the
    final transformation axis, and as for `fftn` for the remaining
    transformation axes.

    See `fft` for details, definitions and conventions used.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.ones((2, 2, 2))
    >>> mt.fft.rfftn(a).execute()
    array([[[ 8.+0.j,  0.+0.j],
            [ 0.+0.j,  0.+0.j]],
           [[ 0.+0.j,  0.+0.j],
            [ 0.+0.j,  0.+0.j]]])

    >>> mt.fft.rfftn(a, axes=(2, 0)).execute()
    array([[[ 4.+0.j,  0.+0.j],
            [ 4.+0.j,  0.+0.j]],
           [[ 0.+0.j,  0.+0.j],
            [ 0.+0.j,  0.+0.j]]])

    """
    a = astensor(a)
    axes = validate_fftn(a, s=s, axes=axes, norm=norm)
    op = TensorRFFTN(shape=s, axes=axes, norm=norm, dtype=np.dtype(np.complex_))
    return op(a)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import Base, Entity, get_output_types
from ...serialize import AnyField, Int8Field, KeyField
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape
from ..operands import DataFrameOperandMixin, DataFrameOperand
from ..utils import parse_index, validate_axis


class DataFrameSetAxis(DataFrameOperand, DataFrameOperandMixin):
    _op_code_ = opcodes.DATAFRAME_SET_AXIS

    _input = KeyField('input')
    _axis = Int8Field('axis')
    _value = AnyField('value')

    def __init__(self, value=None, axis=None, **kw):
        super().__init__(_value=value, _axis=axis, **kw)

    @property
    def input(self):
        return self._input

    @property
    def value(self):
        return self._value

    @property
    def axis(self):
        return self._axis

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = inputs[0]
        if isinstance(self.value, (Base, Entity)):
            self._value = inputs[-1]

    def __call__(self, df_or_series):
        new_size = self.value.shape[0]
        expect_size = df_or_series.axes[self.axis].shape[0]
        if not np.isnan(new_size) and not np.isnan(expect_size) \
                and new_size != expect_size:
            raise ValueError(
                f'Length mismatch: Expected axis has {expect_size} elements, '
                f'new values have {new_size} elements'
            )

        params = df_or_series.params
        if self.axis == 0:
            params['index_value'] = parse_index(self.value) \
                if isinstance(self.value, pd.Index) else self.value.index_value
        else:
            params['columns_value'] = parse_index(self.value, store_data=True) \
                if isinstance(self.value, pd.Index) else self.value.index_value
            pd_columns = self.value.index_value.to_pandas() \
                if isinstance(self.value, (Base, Entity)) else self.value
            params['dtypes'] = params['dtypes'].set_axis(pd_columns)

        self._output_types = get_output_types(df_or_series)
        inputs = [df_or_series]
        if isinstance(self.value, (Base, Entity)):
            inputs += [self.value]
        return self.new_tileable(inputs, **params)

    @classmethod
    def tile(cls, op: 'DataFrameSetAxis'):
        output = op.outputs[0]
        input_tileables = [op.input]

        value = op.value
        if isinstance(value, (Base, Entity)):
            input_tileables.append(value)
            check_chunks_unknown_shape([value], TilesError)

        if any(np.isnan(s) for s in op.input.nsplits[op.axis]):
            raise TilesError(f'chunks has unknown shape on axis {op.axis}')

        if op.input.shape[op.axis] != value.shape[0]:
            raise ValueError(
                f'Length mismatch: Expected axis has {value.shape[0]} elements, '
                f'new values have {op.input.shape[op.axis]} elements'
            )

        if isinstance(value, (Base, Entity)):
            value = value.rechunk({0: op.input.nsplits[op.axis]})._inplace_tile()
            input_tileables[-1] = value

        slices = np.array((0,) + op.input.nsplits[op.axis]).cumsum()
        slice_left = slices[:-1]
        slice_right = slices[1:]

        chunks = []
        param_cache = [None] * len(op.input.nsplits[op.axis])
        for inp_chunk in op.input.chunks:
            input_chunks = [inp_chunk]
            value_index = inp_chunk.index[op.axis]
            params = inp_chunk.params

            if isinstance(value, (Base, Entity)):
                value_data = value.chunks[value_index]
                input_chunks.append(value_data)
            else:
                value_data = value[slice_left[value_index]:slice_right[value_index]]

            if param_cache[value_index] is None:
                cached_params = param_cache[value_index] = dict()
                if isinstance(value, (Base, Entity)):
                    if op.axis == 0:
                        cached_params['index_value'] = value_data.index_value
                    else:
                        cached_params['columns_value'] = value_data.index_value
                        cached_params['dtypes'] = output.dtypes.iloc[
                            slice_left[value_index]:slice_right[value_index]
                        ]
                else:
                    if op.axis == 0:
                        cached_params['index_value'] = parse_index(value_data)
                    else:
                        cached_params['columns_value'] = parse_index(value_data, store_data=True)
                        cached_params['dtypes'] = params['dtypes'].set_axis(value_data)

            params.update(param_cache[value_index])

            new_op = op.copy().reset_key()
            new_op._value = value_data
            chunks.append(new_op.new_chunk(input_chunks, **params))

        params = op.outputs[0].params
        params['chunks'] = chunks
        params['nsplits'] = op.input.nsplits
        new_op = op.copy().reset_key()
        return new_op.new_tileables(input_tileables, **params)

    @classmethod
    def execute(cls, ctx, op: 'DataFrameSetAxis'):
        in_data = ctx[op.input.key]
        value = op.value
        if isinstance(value, (Base, Entity)):
            value = ctx[value.key]
        ctx[op.outputs[0].key] = in_data.set_axis(value, axis=op.axis)


def _set_axis(df_or_axis, labels, axis=0, inplace=False):
    axis = validate_axis(axis, df_or_axis)
    if not isinstance(labels, (Base, Entity)) and not isinstance(labels, pd.Index):
        labels = pd.Index(labels)

    op = DataFrameSetAxis(value=labels, axis=axis)
    result = op(df_or_axis)
    if inplace:
        df_or_axis.data = result.data
    else:
        return result


def df_set_axis(df, labels, axis=0, inplace=False):
    """
    Assign desired index to given axis.

    Indexes for column or row labels can be changed by assigning
    a list-like or Index.

    Parameters
    ----------
    labels : list-like, Index
        The values for the new index.

    axis : {0 or 'index', 1 or 'columns'}, default 0
        The axis to update. The value 0 identifies the rows, and 1 identifies the columns.

    inplace : bool, default False
        Whether to return a new DataFrame instance.

    Returns
    -------
    renamed : DataFrame or None
        An object of type DataFrame or None if ``inplace=True``.

    See Also
    --------
    DataFrame.rename_axis : Alter the name of the index or columns.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})

    Change the row labels.

    >>> df.set_axis(['a', 'b', 'c'], axis='index').execute()
       A  B
    a  1  4
    b  2  5
    c  3  6

    Change the column labels.

    >>> df.set_axis(['I', 'II'], axis='columns').execute()
       I  II
    0  1   4
    1  2   5
    2  3   6

    Now, update the labels inplace.

    >>> df.set_axis(['i', 'ii'], axis='columns', inplace=True)
    >>> df.execute()
       i  ii
    0  1   4
    1  2   5
    2  3   6
    """
    return _set_axis(df, labels, axis=axis, inplace=inplace)


def series_set_axis(series, labels, axis=0, inplace=False):
    """
    Assign desired index to given axis.

    Indexes for row labels can be changed by assigning
    a list-like or Index.

    Parameters
    ----------
    labels : list-like, Index
        The values for the new index.

    axis : {0 or 'index'}, default 0
        The axis to update. The value 0 identifies the rows.

    inplace : bool, default False
        Whether to return a new Series instance.

    Returns
    -------
    renamed : Series or None
        An object of type Series or None if ``inplace=True``.

    See Also
    --------
    Series.rename_axis : Alter the name of the index.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> s = md.Series([1, 2, 3])
    >>> s.execute()
    0    1
    1    2
    2    3
    dtype: int64

    >>> s.set_axis(['a', 'b', 'c'], axis=0).execute()
    a    1
    b    2
    c    3
    dtype: int64
    """
    return _set_axis(series, labels, axis=axis, inplace=inplace)



# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

from ...actors import ActorNotExist, new_client
from .core import ReadinessActor


def main():
    """
    Readiness probe for Mars schedulers and workers
    """
    client = new_client()
    try:
        ref = client.actor_ref(ReadinessActor.default_uid(),
                               address='127.0.0.1:' + os.environ['MARS_K8S_SERVICE_PORT'])
        sys.exit(0 if client.has_actor(ref) else 1)
    except (ActorNotExist, ConnectionRefusedError) as ex:  # noqa: E722
        sys.stderr.write(f'Probe error: {ex}')
        sys.exit(1)


if __name__ == '__main__':   # pragma: no branch
    main()



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType
from ...serialize import KeyField, AnyField, Int32Field, Int64Field
from ...tensor.rechunk.core import get_nsplits, plan_rechunks, compute_rechunk_slices
from ...tensor.utils import calc_sliced_size
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape
from ..initializer import DataFrame as asdataframe, Series as asseries
from ..operands import DataFrameOperand, DataFrameOperandMixin, DATAFRAME_TYPE
from ..utils import indexing_index_value, merge_index_value


class DataFrameRechunk(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.RECHUNK

    _input = KeyField('input')
    _chunk_size = AnyField('chunk_size')
    _threshold = Int32Field('threshold')
    _chunk_size_limit = Int64Field('chunk_size_limit')

    def __init__(self, chunk_size=None, threshold=None, chunk_size_limit=None, output_types=None,
                 reassign_worker=None, **kw):
        super().__init__(_chunk_size=chunk_size, _threshold=threshold,
                         _chunk_size_limit=chunk_size_limit, _output_types=output_types,
                         _reassign_worker=reassign_worker, **kw)

    @property
    def input(self):
        return self._input

    @property
    def chunk_size(self):
        return self._chunk_size

    @property
    def threshold(self):
        return self._threshold

    @property
    def chunk_size_limit(self):
        return self._chunk_size_limit

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, x):
        if isinstance(x, DATAFRAME_TYPE):
            return self.new_dataframe([x], shape=x.shape, dtypes=x.dtypes,
                                      columns_value=x.columns_value, index_value=x.index_value)
        else:
            self.output_types = x.op.output_types
            f = self.new_series if self.output_types[0] == OutputType.series else self.new_index
            return f([x], shape=x.shape, dtype=x.dtype, index_value=x.index_value, name=x.name)

    @classmethod
    def tile(cls, op):
        check_chunks_unknown_shape(op.inputs, TilesError)

        a = op.input
        a = asdataframe(a) if a.ndim == 2 else asseries(a)
        chunk_size = _get_chunk_size(a, op.chunk_size)
        if chunk_size == a.nsplits:
            return [a]

        out = op.outputs[0]
        new_chunk_size = tuple(tuple(s for s in cs if s != 0)
                               for cs in chunk_size)
        if isinstance(out, DATAFRAME_TYPE):
            itemsize = max(getattr(dt, 'itemsize', 8) for dt in out.dtypes)
        else:
            itemsize = out.dtype.itemsize
        steps = plan_rechunks(op.inputs[0], new_chunk_size, itemsize,
                              threshold=op.threshold,
                              chunk_size_limit=op.chunk_size_limit)
        for c in steps:
            out = compute_rechunk(out.inputs[0], c)

        if op.reassign_worker:
            for c in out.chunks:
                c.op._reassign_worker = True

        return [out]


def _get_chunk_size(a, chunk_size):
    if isinstance(a, DATAFRAME_TYPE):
        itemsize = max(getattr(dt, 'itemsize', 8) for dt in a.dtypes)
    else:
        itemsize = a.dtype.itemsize
    return get_nsplits(a, chunk_size, itemsize)


def rechunk(a, chunk_size, threshold=None, chunk_size_limit=None, reassign_worker=False):
    if not any(pd.isna(s) for s in a.shape) and not a.is_coarse():
        try:
            check_chunks_unknown_shape([a], ValueError)
        except ValueError:
            # due to reason that tileable has unknown chunk shape,
            # just ignore to hand over to operand
            pass
        else:
            # do client check only when no unknown shape,
            # real nsplits will be recalculated inside `tile`
            chunk_size = _get_chunk_size(a, chunk_size)
            if chunk_size == a.nsplits:
                return a

    op = DataFrameRechunk(chunk_size=chunk_size, threshold=threshold,
                          chunk_size_limit=chunk_size_limit,
                          reassign_worker=reassign_worker)
    return op(a)


def _concat_dataframe_meta(to_concat_chunks):
    if to_concat_chunks[0].index_value.to_pandas().empty:
        index_value = to_concat_chunks[0].index_value
    else:
        idx_to_index_value = dict((c.index[0], c.index_value) for c in to_concat_chunks if c.index[1] == 0)
        index_value = merge_index_value(idx_to_index_value)

    idx_to_columns_value = dict((c.index[1], c.columns_value) for c in to_concat_chunks if c.index[0] == 0)
    columns_value = merge_index_value(idx_to_columns_value, store_data=True)

    idx_to_dtypes = dict((c.index[1], c.dtypes) for c in to_concat_chunks if c.index[0] == 0)
    dtypes = pd.concat([v[1] for v in list(sorted(idx_to_dtypes.items()))])
    return index_value, columns_value, dtypes


def _concat_series_index(to_concat_chunks):
    if to_concat_chunks[0].index_value.to_pandas().empty:
        index_value = to_concat_chunks[0].index_value
    else:
        idx_to_index_value = dict((c.index[0], c.index_value) for c in to_concat_chunks)
        index_value = merge_index_value(idx_to_index_value)
    return index_value


def compute_rechunk(a, chunk_size):
    from ..indexing.iloc import DataFrameIlocGetItem, SeriesIlocGetItem
    from ..merge.concat import DataFrameConcat

    result_slices = compute_rechunk_slices(a, chunk_size)
    result_chunks = []
    idxes = itertools.product(*[range(len(c)) for c in chunk_size])
    chunk_slices = itertools.product(*result_slices)
    chunk_shapes = itertools.product(*chunk_size)
    is_dataframe = isinstance(a, DATAFRAME_TYPE)
    for idx, chunk_slice, chunk_shape in zip(idxes, chunk_slices, chunk_shapes):
        to_merge = []
        merge_idxes = itertools.product(*[range(len(i)) for i in chunk_slice])
        for merge_idx, index_slices in zip(merge_idxes, itertools.product(*chunk_slice)):
            chunk_index, chunk_slice = zip(*index_slices)
            old_chunk = a.cix[chunk_index]
            merge_chunk_shape = tuple(calc_sliced_size(s, chunk_slice[0]) for s in old_chunk.shape)
            new_index_value = indexing_index_value(old_chunk.index_value, chunk_slice[0])
            if is_dataframe:
                new_columns_value = indexing_index_value(old_chunk.columns_value, chunk_slice[1], store_data=True)
                merge_chunk_op = DataFrameIlocGetItem(chunk_slice, sparse=old_chunk.op.sparse,
                                                      output_types=[OutputType.dataframe])
                merge_chunk = merge_chunk_op.new_chunk([old_chunk], shape=merge_chunk_shape,
                                                       index=merge_idx, index_value=new_index_value,
                                                       columns_value=new_columns_value,
                                                       dtypes=old_chunk.dtypes.iloc[chunk_slice[1]])
            else:
                merge_chunk_op = SeriesIlocGetItem(chunk_slice, sparse=old_chunk.op.sparse,
                                                   output_types=a.op.output_types)
                merge_chunk = merge_chunk_op.new_chunk([old_chunk], shape=merge_chunk_shape,
                                                       index=merge_idx, index_value=new_index_value,
                                                       dtype=old_chunk.dtype, name=old_chunk.name)
            to_merge.append(merge_chunk)
        if len(to_merge) == 1:
            chunk_op = to_merge[0].op.copy()
            if is_dataframe:
                out_chunk = chunk_op.new_chunk(to_merge[0].op.inputs, shape=chunk_shape,
                                               index=idx, index_value=to_merge[0].index_value,
                                               columns_value=to_merge[0].columns_value,
                                               dtypes=to_merge[0].dtypes)
            else:
                out_chunk = chunk_op.new_chunk(to_merge[0].op.inputs, shape=chunk_shape,
                                               index=idx, index_value=to_merge[0].index_value,
                                               name=to_merge[0].name, dtype=to_merge[0].dtype)
            result_chunks.append(out_chunk)
        else:
            if is_dataframe:
                chunk_op = DataFrameConcat(output_types=[OutputType.dataframe])
                index_value, columns_value, dtypes = _concat_dataframe_meta(to_merge)
                out_chunk = chunk_op.new_chunk(to_merge, shape=chunk_shape,
                                               index=idx, index_value=index_value,
                                               columns_value=columns_value,
                                               dtypes=dtypes)
            else:
                chunk_op = DataFrameConcat(output_types=a.op.output_types)
                index_value = _concat_series_index(to_merge)
                out_chunk = chunk_op.new_chunk(to_merge, shape=chunk_shape,
                                               index=idx, index_value=index_value,
                                               dtype=to_merge[0].dtype,
                                               name=to_merge[0].name)
            result_chunks.append(out_chunk)

    if is_dataframe:
        op = DataFrameRechunk(chunk_size, output_types=[OutputType.dataframe])
        return op.new_dataframe([a], a.shape, dtypes=a.dtypes, columns_value=a.columns_value,
                                index_value=a.index_value, nsplits=chunk_size, chunks=result_chunks)
    else:
        op = DataFrameRechunk(chunk_size, output_types=a.op.output_types)
        if a.op.output_types[0] == OutputType.index:
            f = op.new_index
        else:
            f = op.new_series
        return f([a], a.shape, dtype=a.dtype, index_value=a.index_value,
                 nsplits=chunk_size, chunks=result_chunks, name=a.name)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ...actors import FunctionActor
from ...config import options
from ...lib import sparse
from ...serialize import dataserializer
from ..dataio import ArrowBufferIO
from ..utils import WorkerClusterInfoActor
from .core import StorageHandler, ObjectStorageMixin, BytesStorageIO, \
    DataStorageDevice, wrap_promised, register_storage_handler_cls

try:
    import vineyard
    from vineyard._C import ObjectMeta
    from vineyard.core import default_builder_context, default_resolver_context
    from vineyard.data.utils import from_json, to_json
except ImportError:
    vineyard = None
try:
    import pyarrow
except ImportError:
    pyarrow = None


logger = logging.getLogger(__name__)


def mars_sparse_matrix_builder(client, value, builder, **kw):
    meta = ObjectMeta()
    meta['typename'] = 'vineyard::SparseMatrix<%s>' % value.dtype.name
    meta['shape_'] = to_json(value.shape)
    meta.add_member('spmatrix', builder.run(client, value.spmatrix, **kw))
    return client.create_metadata(meta)


def mars_sparse_matrix_resolver(obj, resolver):
    meta = obj.meta
    shape = from_json(meta['shape_'])
    spmatrix = resolver.run(obj.member('spmatrix'))
    return sparse.matrix.SparseMatrix(spmatrix, shape=shape)


if vineyard is not None:
    default_builder_context.register(sparse.matrix.SparseMatrix, mars_sparse_matrix_builder)
    default_resolver_context.register('vineyard::SparseMatrix', mars_sparse_matrix_resolver)


class VineyardKeyMapActor(FunctionActor):
    @classmethod
    def default_uid(cls):
        return 's:0:' + cls.__name__

    def __init__(self):
        super().__init__()
        self._mapping = dict()

    def put(self, session_id, chunk_key, obj_id):
        logger.debug('mapper put: session_id = %s, data_key = %s, data_id = %r', session_id, chunk_key, obj_id)
        session_chunk_key = (session_id, chunk_key)
        target = self._mapping.get(session_chunk_key)
        if target:
            count = target[1] + 1
        else:
            count = 1
        self._mapping[session_chunk_key] = (obj_id, count)

    def get(self, session_id, chunk_key):
        logger.debug('mapper get: session_id = %s, data_key = %s', session_id, chunk_key)
        target = self._mapping.get((session_id, chunk_key))
        if target:
            return target[0]
        else:
            return None

    def batch_get(self, session_id, chunk_keys):
        obj_ids = []
        for key in chunk_keys:
            obj_ids.append(self.get(session_id, key))
        return obj_ids

    def delete(self, session_id, chunk_key):
        session_chunk_key = (session_id, chunk_key)
        target = self._mapping.get(session_chunk_key)
        logger.debug('mapper delete: session_id = %s, chunk_key = %s, target = %s', session_id, chunk_key, target)
        if target:
            if target[1] <= 1:
                self._mapping[session_chunk_key] = (target[0], target[1] - 1)
                return False
            else:
                del self._mapping[session_chunk_key]
                return True
        logger.warn("delete non-existing chunk from mapper: session_id = %s, chunk_key = %s", session_id, chunk_key)
        return True

    def batch_delete(self, session_id, chunk_keys):
        logger.debug('mapper delete: session_id = %s, data_keys = %s', session_id, chunk_keys)
        deletable = []
        for key in chunk_keys:
            deletable.append(self.delete(session_id, key))
        return deletable


class VineyardBytesIO(BytesStorageIO):
    storage_type = DataStorageDevice.VINEYARD

    def __init__(self, vineyard_client, session_id, data_key, data_id, mode='w',
                 nbytes=None, packed=False, compress=None, auto_register=True,
                 pin_token=None, handler=None):
        from .objectholder import SharedHolderActor

        logger.debug('create vineyard bytes IO: mode = %s, packed = %s, compress = %r',
                     mode, packed, compress)

        super().__init__(session_id, data_key, mode=mode, handler=handler)
        self._client = vineyard_client
        self._data_id = data_id
        self._buffer = None
        self._offset = 0
        self._nbytes = nbytes
        self._holder_ref = self._storage_ctx.actor_ctx.actor_ref(SharedHolderActor.default_uid())
        self._compress = compress or dataserializer.CompressType.NONE
        self._packed = packed
        self._auto_register = auto_register
        self._pin_token = pin_token

        block_size = options.worker.copy_block_size

        if self.is_writable:
            logger.debug('bytes io write: session_id = %s, data_key = %s, size = %d',
                         session_id, data_key, nbytes)
            self._buffer = pyarrow.allocate_buffer(nbytes, resizable=False)
            if packed:
                self._buf = ArrowBufferIO(self._buffer, 'w', block_size=block_size)
            else:
                self._buf = pyarrow.FixedSizeBufferWriter(self._buffer)
                self._buf.set_memcopy_threads(6)
        elif self.is_readable:
            logger.debug('bytes io get: session_id = %s, data_key = %s, data_id = %r',
                         session_id, data_key, data_id)
            data = self._client.get(data_id)

            self._buffer = pyarrow.serialize(data, dataserializer.mars_serialize_context()).to_buffer()
            if packed:
                self._buf = ArrowBufferIO(
                    self._buffer, 'r', compress_out=compress, block_size=block_size)
                self._nbytes = len(self._buffer)
            else:
                self._mv = memoryview(self._buffer)
                self._nbytes = len(self._buffer)
        else:
            raise NotImplementedError

    def __del__(self):
        self._buf = self._buffer = None

    @property
    def nbytes(self):
        return self._nbytes

    def read(self, size=-1):
        if self._packed:
            return self._buf.read(size)
        else:
            if size < 0:
                size = self._nbytes
            right_pos = min(self._nbytes, self._offset + size)
            ret = self._mv[self._offset:right_pos]
            self._offset = right_pos
            return ret

    def write(self, d):
        return self._buf.write(d)

    def close(self, finished=True):
        if self._closed:
            return

        if self.is_writable and self._buffer is not None:
            if finished:
                data = pyarrow.deserialize(self._buffer, dataserializer.mars_serialize_context())
                if hasattr(data, 'shape'):
                    data_shape = getattr(data, 'shape')
                else:
                    data_shape = (1,)
                self._handler.put_objects(self._session_id, [self._data_key], [data],
                                          sizes=[self._nbytes], shapes=[data_shape])

        self._buf = self._buffer = None
        super().close(finished=finished)


class VineyardHandler(StorageHandler, ObjectStorageMixin):
    storage_type = DataStorageDevice.VINEYARD

    def __init__(self, storage_ctx, proc_id=None):
        StorageHandler.__init__(self, storage_ctx, proc_id=proc_id)
        self._client = vineyard.connect(options.vineyard.socket)
        self._cluster_info = self._actor_ctx.actor_ref(WorkerClusterInfoActor.default_uid())

    def _new_object_id(self, session_id, data_key, data_id):
        addr = self._cluster_info.get_scheduler((session_id, data_key))
        return self._actor_ctx.actor_ref(VineyardKeyMapActor.default_uid(), address=addr) \
            .put(session_id, data_key, data_id)

    def _get_object_id(self, session_id, data_key):
        addr = self._cluster_info.get_scheduler((session_id, data_key))
        obj_id = self._actor_ctx.actor_ref(VineyardKeyMapActor.default_uid(), address=addr) \
            .get(session_id, data_key)
        return obj_id

    def _batch_get_object_id(self, session_id, data_keys):
        obj_ids = []
        for data_key in data_keys:
            addr = self._cluster_info.get_scheduler((session_id, data_key))
            obj_id = self._actor_ctx.actor_ref(VineyardKeyMapActor.default_uid(), address=addr) \
                .get(session_id, data_key)
            obj_ids.append(obj_id)
        return obj_ids

    def _batch_delete_from_key_mapper(self, session_id, data_keys):
        deletable = []
        for data_key in data_keys:
            addr = self._cluster_info.get_scheduler((session_id, data_key))
            val = self._actor_ctx.actor_ref(VineyardKeyMapActor.default_uid(), address=addr) \
                .delete(session_id, data_key)
            deletable.append(val)
        return deletable

    @wrap_promised
    def create_bytes_reader(self, session_id, data_key, packed=False, packed_compression=None,
                            _promise=False):
        data_id = self._get_object_id(session_id, data_key)
        logger.debug('create vineyard bytes reader: data_id = %s', data_id)
        return VineyardBytesIO(self._client, session_id, data_key, data_id, 'r', packed=packed,
                               compress=packed_compression, handler=self)

    @wrap_promised
    def create_bytes_writer(self, session_id, data_key, total_bytes, packed=False,
                            packed_compression=None, auto_register=True, pin_token=None,
                            _promise=False):
        logger.debug('create vineyard bytes writer: data_key = %s', data_key)
        return VineyardBytesIO(self._client, session_id, data_key, 'w', nbytes=total_bytes,
                               packed=packed, compress=packed_compression, auto_register=auto_register,handler=self)

    @wrap_promised
    def get_objects(self, session_id, data_keys, serialize=False, _promise=False):
        data_ids = [self._get_object_id(session_id, data_key) for data_key in data_keys]
        return [self._client.get(data_id) for data_id in data_ids]

    @wrap_promised
    def put_objects(self, session_id, data_keys, objs, sizes=None, shapes=None,
                    serialize=False, pin_token=None, _promise=False):
        new_sizes = []
        for data_key, obj in zip(data_keys, objs):
            if isinstance(obj, pyarrow.SerializedPyObject):
                obj = obj.deserialize(dataserializer.mars_serialize_context())
            data_id = self._client.put(obj)
            self._new_object_id(session_id, data_key, data_id)
            # FIXME its is a hack to fixes the nbytes mismatch in bytes reader/writer.
            #
            # refresh the data size attribute for bytes reader.
            new_sizes.append(pyarrow.serialize(self._client.get(data_id),
                                               dataserializer.mars_serialize_context()).total_bytes)
        self.register_data(session_id, data_keys, new_sizes, shapes)

    def load_from_bytes_io(self, session_id, data_keys, src_handler, pin_token=None):
        def _read_serialized(reader):
            with reader:
                return reader.get_io_pool().submit(reader.read).result()

        def _fallback(*_):
            return self._batch_load_objects(
                session_id, data_keys,
                lambda k: src_handler.create_bytes_reader(session_id, k, _promise=True).then(_read_serialized),
                serialize=True
            )

        return self.transfer_in_runner(session_id, data_keys, src_handler, _fallback)

    def load_from_object_io(self, session_id, data_keys, src_handler, pin_token=None):
        def _fallback(*_):
            return self._batch_load_objects(
                session_id, data_keys,
                lambda k: src_handler.get_objects(session_id, k, _promise=True), batch_get=True)

        return self.transfer_in_runner(session_id, data_keys, src_handler, _fallback)

    def delete(self, session_id, data_keys, _tell=False):
        data_ids = self._batch_get_object_id(session_id, data_keys)
        logger.debug('delete chunks from vineyard: keys = %s, in vineyard is %s', data_keys, data_ids)
        # when multiple workers connected to the same vineyard, they will think they both
        # hold the chunk, but when it being deleted, it won't be there anymore.
        data_ids = [data_id for data_id in data_ids if data_id]

        if data_ids:
            deletable = self._batch_delete_from_key_mapper(session_id, data_keys)
            data_ids_to_delete = [data_id for data_id, val in zip(data_ids, deletable)
                                          if val]
            try:
                self._client.delete(data_ids_to_delete, deep=True)
            except vineyard._C.ObjectNotExistsException:
                # the object may has been deleted by other worker
                pass
        self.unregister_data(session_id, data_keys, _tell=_tell)


register_storage_handler_cls(DataStorageDevice.VINEYARD, VineyardHandler)
